
/*
Hypothesis: With increasing allocation rates and number threads, the fraction of program
time spend on GC will seriously hinder program performance/scalablity.
 
Basically this benchmark is a linked list of linked lists of numbers.  (A nested list to promote 
discontinous memory usage.)  The operation is a full traversal and at each node there is a chance of
(1) writing, to consider if update logs are less space efficient than read logs in some systems
(2) allocation immediately eligible for collection
(3) allocation eligible for collection after commit
(4) insertion = new object that will survive after commit
(5) deletion = old object now eligible for collection
So those + size of the list gives 6 parameters we can vary continuously.

This benchmark has no contention, which means it will not reveal the extra gc pressure from 
aborted transactions.

JVM flags
-verbosegc
-Xmx512M
-XX:+PrintGCDetails

http://www.oracle.com/technetwork/java/gc-tuning-5-138395.html
*/
public class GCBenchmark
{
	final double writeRatio, immediateRatio, collectRatio, insertRatio, removeRatio;
	final int size, nthreads; 
	
	private GCBenchmark(double writeRatio, double immediateRatio, double collectRatio, double insertRatio,
		double removeRatio, int size, int nthreads)
	{
		this.writeRatio = writeRatio;
		this.immediateRatio = immediateRatio;
		this.collectRatio = collectRatio;
		this.insertRatio = insertRatio;
		this.removeRatio = removeRatio;
		this.size = size;
		this.nthreads = nthreads;
	}
	
	class BenchmarkThread extends Thread
	{
		Node root;
		
		private BenchmarkThread()
		{
			root = null;
			
			for(int i=0; i<size; i++)
			{
				root = new Node( newSubList(), root );
			}
		}
	
		private Node newSubList()
		{
			Node n = null;
			
			for(int i=0; i<size; i++)
			{
				n = new Node( new Box(random()), n);
			}
			
			return n;
		}
		
		private int seed = 13371337;
		private double random()
		{
			seed = (( seed * 1664525 ) + 1013904223) % Integer.MAX_VALUE;
			return ((double)Math.abs(seed)) / ((double)Integer.MAX_VALUE);
		}
		
		public void run()
		{
			try
			{
				for(int i=0; i<25; i++) transaction();
			}
			catch(Throwable t)
			{
				t.printStackTrace();
				System.exit(-1);
			}
		}
	
		@org.multiverse.annotations.TransactionalMethod
		private void transaction()
		{
			Node collect = null;
			Node n = root;
			while(n != null)
			{
				Node m = (Node)n.value;
				while(m != null)
				{
					double v = ((Box)m.value).value; //always read
					
					if( random() < writeRatio ) //include some fraction of writes in case update logs much less space efficient than read logs
					{
						m.value = new Box( random() );
					}
					if( random() < immediateRatio ) //some fraction of allocations that are immediately eligable for garbage collection
					{
						new Node( new Box(0), m );
					}	
					if( random() < collectRatio ) //some fraction of allocations that will be eligable for garbage collection after commit
					{
						collect = new Node( new Box(v), collect );
					}
					if( random() < insertRatio ) //some fraction of insertion: new values that will survive
					{
						Node ins = new Node( new Box(random()), m.next );
						m.next = ins;
					}
					if( random() < removeRatio ) //some fraction of removal: old values that will be collected
					{
						if(m.next != null) m.next = m.next.next;
					}
					
					m = m.next;
				}
				
				if( random() < immediateRatio ) //some fraction of allocations that are immediately eligable for garbage collection
				{
					new Node( new Box(0), n );
				}		
				if( random() < collectRatio ) //some fraction of allocations that will be eligable for garbage collection after commit
				{
					collect = new Node( n, collect );
				}	
				if( random() < insertRatio ) //some fraction of insertion: new values that will survive
				{
					Node ins = new Node( newSubList(), n.next );
					n.next = ins;
				}
				if( random() < removeRatio ) //some fraction of removal: old values that will be collected
				{
					if(n.next != null) n.next = n.next.next;
				}
				
				n = n.next;
			}
		}
	}
	
	public void main(boolean warmup) throws InterruptedException
	{
		if(!warmup){
			System.out.println("WRITE\t"+writeRatio);
			System.out.println("IMMED\t"+immediateRatio);
			System.out.println("COLLECT\t"+collectRatio);
			System.out.println("INSERT\t"+insertRatio);
			System.out.println("REMOVE\t"+removeRatio);
			System.out.println("SIZE\t"+size);
			System.out.println("THREADS\t"+nthreads);
		}
	
		BenchmarkThread[] threads = new BenchmarkThread[nthreads];
		for(int i=0;i<nthreads;i++) threads[i] = new BenchmarkThread();
		
		if(!warmup) System.out.println("Initialized");
		
		long start = System.currentTimeMillis();
		
		for(int i=0;i<nthreads;i++) threads[i].start();
		for(int i=0;i<nthreads;i++) threads[i].join();
		
		long duration = System.currentTimeMillis() - start;
		if(!warmup) System.out.println("Duration\t"+duration);
	}

	public static void main(String[] args) throws InterruptedException
	{
		//default values
		double writeRatio = 0.05;
		double immediateRatio = 0.15;
		double collectRatio = 0.65;
		double insertRatio = 0.05;
		double removeRatio = 0.05;
		int size = 200; //of both the list and sublists
		int nthreads = 2;
		
		for(int i=0; i<args.length; i++)
		{
			if(args[i].startsWith("write=")) writeRatio = Double.valueOf( args[i].substring(6) );
			if(args[i].startsWith("immediate=")) immediateRatio = Double.valueOf( args[i].substring(10) );
			if(args[i].startsWith("collect=")) collectRatio = Double.valueOf( args[i].substring(8) );
			if(args[i].startsWith("insert=")) insertRatio = Double.valueOf( args[i].substring(7) );
			if(args[i].startsWith("remove=")) removeRatio = Double.valueOf( args[i].substring(7) );
			if(args[i].startsWith("size=")) size = Integer.valueOf( args[i].substring(5) );
			if(args[i].startsWith("threads=")) nthreads = Integer.valueOf( args[i].substring(8) );
		}
	
		new GCBenchmark( writeRatio, immediateRatio, collectRatio, insertRatio, removeRatio, 200, nthreads ).main(true);
		new GCBenchmark( writeRatio, immediateRatio, collectRatio, insertRatio, removeRatio, 200, nthreads ).main(true);
		System.gc();
		new GCBenchmark( writeRatio, immediateRatio, collectRatio, insertRatio, removeRatio, size, nthreads ).main(false);
	}

}


@org.multiverse.annotations.TransactionalObject
class Node
{
	public Object value;
	public Node next;
	
	public Node(Object value, Node next)
	{
		this.value = value;
		this.next = next;
	}
}

@org.multiverse.annotations.TransactionalObject
class Box
{
	public double value;
	public Box(double value)
	{
		this.value = value;
	}
}
