
/*
Hypothesis: With increasing allocation rates and number threads, the fraction of program
time spend on GC will seriously hinder program performance/scalablity.
 
Basically this benchmark is a linked list of linked lists of numbers.  (A nested list to promote 
discontinous memory usage.)  The operation is a full traversal and at each node there is a chance of
(1) writing, to consider if update logs are less space efficient than read logs in some systems
(2) allocation immediately eligible for collection
(3) allocation eligible for collection after commit
(4) insertion = new object that will survive after commit
(5) deletion = old object now eligible for collection
So those + size of the list gives 6 parameters we can vary continuously.

This benchmark has no contention, which means it will not reveal the extra gc pressure from 
aborted transactions.

JVM flags
-verbosegc
-Xmx512M
-XX:+PrintGCDetails

http://www.oracle.com/technetwork/java/gc-tuning-5-138395.html
*/
using System;
using System.Threading;

public class GCBenchmark
{
	static double writeRatio, immediateRatio, collectRatio, insertRatio, removeRatio;
	static int size, nthreads; 
	
	private GCBenchmark(double writeRatio, double immediateRatio, double collectRatio, double insertRatio,
		double removeRatio, int size, int nthreads)
	{
		GCBenchmark.writeRatio = writeRatio;
        GCBenchmark.immediateRatio = immediateRatio;
        GCBenchmark.collectRatio = collectRatio;
        GCBenchmark.insertRatio = insertRatio;
        GCBenchmark.removeRatio = removeRatio;
        GCBenchmark.size = size;
        GCBenchmark.nthreads = nthreads;
	}
	
	public class BenchmarkThread
	{
		Node root;
		
		public BenchmarkThread()
		{
			root = null;
			
			for(int i=0; i<size; i++)
			{
				root = new Node( newSubList(), root );
			}
		}
	
		private Node newSubList()
		{
			Node n = null;
			
			for(int i=0; i<size; i++)
			{
				n = new Node( new Box(random()), n);
			}
			
			return n;
		}
		
		private int seed = 13371337;
		private double random()
		{
            seed = ((seed * 1664525) + 1013904223) % int.MaxValue;
			return ((double)Math.Abs(seed)) / ((double)int.MaxValue);
		}
		
		public void run()
		{
			for(int i=0; i<25; i++)	transaction();
		}
	
		private void transaction()
		{
			Node collect = null;
			Node n = root;
			while(n != null)
			{
				Node m = (Node)n.value;
				while(m != null)
				{
					double v = ((Box)m.value).value; //always read
					
					if( random() < writeRatio ) //include some fraction of writes in case update logs much less space efficient than read logs
					{
						m.value = new Box( random() );
					}
					if( random() < immediateRatio ) //some fraction of allocations that are immediately eligable for garbage collection
					{
						new Node( new Box(0), m );
					}	
					if( random() < collectRatio ) //some fraction of allocations that will be eligable for garbage collection after commit
					{
						collect = new Node( new Box(v), collect );
					}
					if( random() < insertRatio ) //some fraction of insertion: new values that will survive
					{
						Node ins = new Node( new Box(random()), m.next );
						m.next = ins;
					}
					if( random() < removeRatio ) //some fraction of removal: old values that will be collected
					{
						if(m.next != null) m.next = m.next.next;
					}
					
					m = m.next;
				}
				
				if( random() < immediateRatio ) //some fraction of allocations that are immediately eligable for garbage collection
				{
					new Node( new Box(0), n );
				}		
				if( random() < collectRatio ) //some fraction of allocations that will be eligable for garbage collection after commit
				{
					collect = new Node( n, collect );
				}	
				if( random() < insertRatio ) //some fraction of insertion: new values that will survive
				{
					Node ins = new Node( newSubList(), n.next );
					n.next = ins;
				}
				if( random() < removeRatio ) //some fraction of removal: old values that will be collected
				{
					if(n.next != null) n.next = n.next.next;
				}
				
				n = n.next;
			}
		}
	}
	
	public void main()
	{
		Console.WriteLine("WRITE\t"+writeRatio);
		Console.WriteLine("IMMED\t"+immediateRatio);
		Console.WriteLine("COLLECT\t"+collectRatio);
		Console.WriteLine("INSERT\t"+insertRatio);
		Console.WriteLine("REMOVE\t"+removeRatio);
		Console.WriteLine("SIZE\t"+size);
		Console.WriteLine("THREADS\t"+nthreads);
	
		BenchmarkThread[] benchmarkThreads = new BenchmarkThread[nthreads];
		Thread[] threads = new Thread[nthreads];
        for (int i = 0; i < nthreads; i++) { 
            benchmarkThreads[i] = new BenchmarkThread();
            threads[i] = new Thread(new ThreadStart(benchmarkThreads[i].run));
        }
	

		Console.WriteLine("Initialized");

        long start = DateTime.Now.Ticks;
		
		for(int i=0;i<nthreads;i++) threads[i].Start();
		for(int i=0;i<nthreads;i++) threads[i].Join();

        long duration = DateTime.Now.Ticks - start;
		Console.WriteLine("Duration\t" + Math.Round(new TimeSpan(duration).TotalMilliseconds));
	}

	static void Main(string[] args)
	{
		//default values
		double writeRatio = 0.05;
		double immediateRatio = 0.15;
		double collectRatio = 0.65;
		double insertRatio = 0.05;
		double removeRatio = 0.05;
		int size = 200; //of both the list and sublists
		int nthreads = 2;
		
		for(int i=0; i<args.Length; i++)
		{
			if(args[i].StartsWith("write=")) writeRatio = double.Parse( args[i].Substring(6) );
			if(args[i].StartsWith("immediate=")) immediateRatio = double.Parse( args[i].Substring(10) );
			if(args[i].StartsWith("collect=")) collectRatio = double.Parse( args[i].Substring(8) );
			if(args[i].StartsWith("insert=")) insertRatio = double.Parse( args[i].Substring(7) );
			if(args[i].StartsWith("remove=")) removeRatio = double.Parse( args[i].Substring(7) );
            if (args[i].StartsWith("size=")) size = int.Parse(args[i].Substring(5));
			if(args[i].StartsWith("threads=")) nthreads = int.Parse( args[i].Substring(8) );
		}
	
		GC.Collect();
		new GCBenchmark( writeRatio, immediateRatio, collectRatio, insertRatio, removeRatio, size, nthreads ).main();
	}
}

public class Node
{
	public Object value;
	public Node next;
	
	public Node(Object value, Node next)
	{
		this.value = value;
		this.next = next;
	}
}

public class Box
{
	public double value;
	public Box(double value)
	{
		this.value = value;
	}
}
